local E, L, V, P, G, _ = unpack(ElvUI);
local LP = E:GetModule('LvPlus');
local CB = E:NewModule('ChatBar', 'AceEvent-3.0', 'AceHook-3.0');
local CE = E:GetModule('ChatEmote');
local CSR = E:GetModule('ChatStatReport');
local CBN = E:GetModule('ChatBossNotes');
local CH = E:GetModule('Chat');

local sub, gsub = string.sub, gsub;
local GetChannelName = GetChannelName;
local sort, tinsert = table.sort, table.insert;
local CreateFrame = CreateFrame;
local IsInGuild = IsInGuild;
local CanEditOfficerNote = CanEditOfficerNote;
local IsPartyLFG = IsPartyLFG;
local IsInGroup = IsInGroup;
local IsInRaid = IsInRaid;
local ChatFrame_OpenChat = ChatFrame_OpenChat;
local C_Club_GetClubInfo = C_Club.GetClubInfo;
local unpack = unpack;

local Button = {};

local ChannelList = {
	[L["BigFootChannel"]] = sub(L["BigFootChannel"],7,9),
}

local BlackChannelName = {
	['EUIGVC'] = true,
	['ElvUIGVC'] = true,
}

local function ChannelShortText(index)
	local channelNum, channelName = GetChannelName(index);
	if not channelName or BlackChannelName[channelName] then return 0 end

	if strfind(channelName, 'Community') then
		local _, clubId, _ = strsplit(':', channelName);
		local clubInfo = C_Club_GetClubInfo(clubId);
		if clubInfo then
			channelName = clubInfo.shortName or clubInfo.name;
		end
		return E:ShortenString(gsub(channelName, "%s%-%s.*", ""), 1);
	else
		if channelNum ~= 0 and channelName ~= '集合石' then
			if channelName == L["BigFootChannel"] then
				return ChannelList[L["BigFootChannel"]];
			else
				return E:ShortenString(gsub(channelName, "%s%-%s.*", ""), 1);
			end
		else
			return 0;
		end
	end
end

local ChatTypesList = {
	['CHANNEL1'] = 1,
	['CHANNEL2'] = 2,
	['CHANNEL3'] = 3,
	['CHANNEL4'] = 4,
	['CHANNEL5'] = 5,
	['CHANNEL6'] = 6,
	['CHANNEL7'] = 7,
	['CHANNEL8'] = 8,
	['CHANNEL9'] = 9,	
	['CHANNEL10'] = 10,
	['SAY'] = {
		['name'] = L['says'] or 'S',
		['cmd'] = '/s ',
		['show'] = function() return true; end,
	},
	['YELL'] = {
		['name'] = L['yells'] or 'Y',
		['cmd'] = '/y ',
		['show'] = function() return true; end,
	},
	['GUILD'] = {
		['name'] = L['G'],
		['cmd'] = '/g ',
		['show'] = function() return IsInGuild() end,
	},
	['OFFICER'] = {
		['name'] = L['O'],
		['cmd'] = '/o ',
		['show'] = function() return CanEditOfficerNote() end,
	},
	['PARTY'] = {
		['name'] = L['P'],
		['cmd'] = '/p ',
		['show'] = function() return IsPartyLFG() or IsInGroup() or IsInRaid(); end,
	},
	['RAID'] = {
		['name'] = L['R'],
		['cmd'] = '/raid ',
		['show'] = function() return IsInRaid(); end,
	},
	['RAID_WARNING'] = {
		['name'] = L['RW'],
		['cmd'] = '/rw ',
		['show'] = function() return IsInRaid(); end,
	},
	['INSTANCE_CHAT'] = {
		['name'] = L['I'],
		['cmd'] = '/i ',
		['show'] = function() return IsPartyLFG() end,
	},
	['BATTLE_CHAT'] = {
		['name'] = L['BG'],
		['cmd'] = '/bg ',
		['show'] = function() local inInstance, instanceType = IsInInstance();local isInPVP = inInstance and (instanceType == "pvp"); return isInPVP end,
	},
}
local chatTypesListTable = {}
for k, v in pairs(ChatTypesList) do
	tinsert(chatTypesListTable, k);
end
sort(chatTypesListTable, function(a, b)
	return a < b;
end)

local function ChannelButton(parent, width, height, postion, order, text, color)
	local f = CreateFrame("Button", nil, parent);
	f:Width(width);
	f:Height(height);
	f:SetPoint("LEFT", parent ,"LEFT", (postion == 1) and 7 or (7 + (postion - 1) * 20), 0);
	f:RegisterForClicks("AnyUp");
	f:SetScript("OnClick", function()
		local text = '';
		if ChatFrame1EditBox:IsShown() then
			text = ChatFrame1EditBox:GetText();
		end
		if not order:find(' ') then order = order..' ' end
		ChatFrame_OpenChat(order..text, SELECTED_DOCK_FRAME);
	end)
	
	f.text = f:CreateFontString(nil, 'OVERLAY');
	f.text:FontTemplate(nil, 14, nil);
	f.text:SetText(text);
	f.text:SetPoint("CENTER", 0, 0);
	f.text:SetTextColor(unpack(color));
	
	f:SetScript("OnEnter", function(self)
		self.text:SetShadowColor(0.9, 0, 0, 0.8);
	end)
	f:SetScript("OnLeave", function(self)
		self.text:SetShadowColor(0, 0, 0, 0.2);
	end)
	
	f:Show();
	
	tinsert(Button, f);
end

local function GetChannelColor(name)
	if name and ChatTypeInfo[name] then
		return {ChatTypeInfo[name].r, ChatTypeInfo[name].g, ChatTypeInfo[name].b};
	else
		return {1,1,1};
	end
end

local function CreateChannelButton(chat)
	for i = 1, #Button do
		Button[i]:Hide();
		Button[i] = nil;
	end
	wipe(Button);

	local i = 1;

	for n = 1, #chatTypesListTable do
		if chatTypesListTable[n]:find('CHANNEL') and ChannelShortText(ChatTypesList[chatTypesListTable[n]]) ~= 0 then
			ChannelButton(chat, 16, 16, i, "/"..string.sub(chatTypesListTable[n],8)..' ', ChannelShortText(ChatTypesList[chatTypesListTable[n]]), GetChannelColor(chatTypesListTable[n]));
			i = i + 1;
		end
		if not chatTypesListTable[n]:find('CHANNEL') then
			if ChatTypesList[chatTypesListTable[n]].show() then
				ChannelButton(chat, 16, 16, i, ChatTypesList[chatTypesListTable[n]].cmd, ChatTypesList[chatTypesListTable[n]].name, GetChannelColor(chatTypesListTable[n]));
				i = i + 1;
			end
		end
	end
end

function CB:LoadChatMSGLoot()
	local pattern = "^" .. string.format(LOOT_ITEM, "(.+)", "(.+)") .. "$";
	local slot_name = {
		INVTYPE_TRINKET = "饰品",
		INVTYPE_FINGER = "戒指",
		INVTYPE_CLOAK = "披风",
	};
	ChatFrame_AddMessageEventFilter("CHAT_MSG_LOOT", function(self, event, msg, ...)
		local _, _, player, item = msg:find(pattern);
		if player and player ~= YOU then
			local _, _, quality, _, _, _, _, _, slot = GetItemInfo(item);
			if quality >= 3 and slot and slot ~= "" then
				local _, class = UnitClass(player);
				local color = class and RAID_CLASS_COLORS[class].colorStr;
				local playerText = player;
				if E.db.LvPlus.LvGeneral.General.LvChatFrame.ChatMSGLootGS and TeamStats and TeamStats.db then
					local fullname = player:find("-") and player or player .. "-" .. GetRealmName();
					local data = TeamStats.db.players[fullname];
					if data.gsGot and data.gs then
						playerText = playerText .. "(" .. data.gs .. ")";
					end
				end
				player = format("|Hplayer:%s|h%s[%s]%s|h", player, color and "|c" .. color or "", playerText, color and "|r" or "");
				local msg2 = string.format(LOOT_ITEM, player, item);
				return false, msg2, ...;
			end
		end
	end)
end

function CB:ChatBub()
	if IsInInstance() then
		SetCVar("chatBubbles", 1);
		if E.db.LvPlus.LvGeneral.General.LvChatFrame.SmartChatBub.SmartChatBubTip then
			LP:Print(ENABLE..L["ChatBub"]);
		end
	else
		SetCVar("chatBubbles", 0);
		if E.db.LvPlus.LvGeneral.General.LvChatFrame.SmartChatBub.SmartChatBubTip then
			LP:Print(DISABLE..L["ChatBub"]);
		end
	end
end

function CB:PositionChatbar()
	local NUM_CHAT_WINDOWS = NUM_CHAT_WINDOWS;
	local BASE_OFFSET = 57 + E.Spacing * 3;
	for _, frameName in pairs(_G.CHAT_FRAMES) do
		local frame = _G[frameName];
		local id = frame:GetID();
		if id ~= 2 and not (id > NUM_CHAT_WINDOWS) and id ~= frame.RightChatWindowID then
			frame:SetSize(E.db.chat.panelWidth - 11, (E.db.chat.panelHeight - BASE_OFFSET) - 28);
		end
	end
end

function CB:LoadChatbar()
	local chat = CreateFrame("Frame", "LvChatbarFrame", LeftChatPanel);
	if E.db.LvPlus.LvGeneral.General.LvChatFrame.ChatBar.ChatBarPoi == 0 then
		chat:Point("TOP", LeftChatPanel, "TOP", -3, -1);
	elseif E.db.LvPlus.LvGeneral.General.LvChatFrame.ChatBar.ChatBarPoi == 1 then
		chat:Point("TOP", LeftChatPanel, "TOP", -3, 24);
	else
		chat:Point("TOP", LeftChatPanel, "BOTTOM", -3, 1);
	end
	chat:Width(E.db["chat"].panelWidth);
	chat:Height(25);
	CB.chatbar = chat;

	chat:RegisterEvent("PLAYER_ENTERING_WORLD");
	chat:RegisterEvent("CHAT_MSG_CHANNEL_NOTICE");
	chat:RegisterEvent("UPDATE_CHAT_COLOR");
	chat:RegisterEvent("RAID_ROSTER_UPDATE");
	chat:RegisterEvent("GROUP_ROSTER_UPDATE");
	chat:RegisterEvent("PLAYER_GUILD_UPDATE");
	chat:RegisterEvent("CLUB_ADDED");
	chat:RegisterEvent("CLUB_REMOVED");
	chat:SetScript("OnEvent", function(self)
		E:ScheduleTimer(CreateChannelButton, 0.5, chat);
	end)

	local big = CreateFrame("Button", "Bigfootcannel", self.chatbar);
	big:Point("RIGHT", self.chatbar, "RIGHT", 0, 0);
	big:Size(20);
	big:RegisterForClicks("AnyDown");
	big:SetNormalTexture("Interface\\Addons\\ElvUI_LvPlus\\Media\\world");
	big:SetHighlightTexture("Interface\\Addons\\ElvUI_LvPlus\\Media\\world-Highlight");
	if E.db.LvPlus.LvGeneral.General.LvChatFrame.ChatBar.SyncBtn then
		big:GetNormalTexture():SetDesaturated(false);
	else
		big:GetNormalTexture():SetDesaturated(true);
	end
	big:SetScript("OnClick", function(self)
		if E.db.LvPlus.LvGeneral.General.LvChatFrame.ChatBar.SyncBtn ~= true then
			JoinTemporaryChannel(L["BigFootChannel"]);
			ChatFrame_RemoveChannel(DEFAULT_CHAT_FRAME, L["BigFootChannel"]);
			ChatFrame_AddChannel(DEFAULT_CHAT_FRAME, L["BigFootChannel"]);
			E.db.LvPlus.LvGeneral.General.LvChatFrame.ChatBar.SyncBtn = true;
			self:GetNormalTexture():SetDesaturated(false);
			LP:Print(ENABLE..L["BigFootChannel"]);
		else
			SlashCmdList["LEAVE"](L["BigFootChannel"]);
			E.db.LvPlus.LvGeneral.General.LvChatFrame.ChatBar.SyncBtn = false;
			self:GetNormalTexture():SetDesaturated(true);
			LP:Print(DISABLE..L["BigFootChannel"]);
		end
	end)
	big:SetScript("OnEnter", function(self)
		GameTooltip:SetOwner(self, 'ANCHOR_TOP', 0, 6);
		GameTooltip:AddLine(L["Enable/DisableBigFootChannel"]);
		GameTooltip:Show();
	end)
	big:SetScript("OnLeave", function(self)
		GameTooltip:Hide();
	end)
	
	local roll = CreateFrame("Button", "rollMacro", E.UIParent, "SecureActionButtonTemplate");
	roll:SetAttribute("*type*", "macro");
	roll:SetAttribute("macrotext", "/roll");
	roll:SetParent(self.chatbar);
	roll:Width(18);
	roll:Height(18);
	roll:SetPoint("RIGHT",self.chatbar,"RIGHT", -25,0);
	roll:SetNormalTexture("Interface\\Buttons\\UI-GroupLoot-Dice-Up");
	roll:SetPushedTexture("Interface\\Buttons\\UI-GroupLoot-Dice-Down");
	roll:SetHighlightTexture("Interface\\Buttons\\UI-GroupLoot-Dice-Highlight");
	roll:SetScript("OnEnter", function(self)
		GameTooltip:SetOwner(self, 'ANCHOR_TOP', 0, 6);
		GameTooltip:AddLine(L["Roll 1-100"]);
		GameTooltip:Show();
	end)
	roll:SetScript("OnLeave", function(self)
		GameTooltip:Hide();
	end)
	
	local Emote = CE.ChatEmote;
	local ChatEmote = CreateFrame("Button", nil, self.chatbar);
	ChatEmote:SetPoint("RIGHT",self.chatbar,"RIGHT", -50,0);
	ChatEmote:Size(16);
	ChatEmote:SetScript("OnClick", function() Emote.ToggleEmoteTable() end);
	
	ChatEmote:SetNormalTexture("Interface\\Addons\\ElvUI_LvPlus\\Media\\Smile");
	ChatEmote:SetHighlightTexture("Interface\\Addons\\ElvUI_LvPlus\\Media\\Smile-Highlight");
	ChatEmote:SetScript("OnEnter", function(self)
		GameTooltip:SetOwner(self, 'ANCHOR_TOP', 0, 6);
		GameTooltip:AddLine(Emote.tipstr);
		GameTooltip:Show();
	end);
	ChatEmote:SetScript("OnLeave", function(self)
		GameTooltip:Hide();
	end);


	local statreport = CreateFrame("Button", nil, self.chatbar);
	statreport:Size(16);
	statreport:SetPoint("RIGHT",self.chatbar,"RIGHT", -75,0);
	statreport:RegisterForClicks("AnyUp");
	statreport:SetScript("OnClick", function(self, btn)
		if btn == "RightButton" then
			CBN:SendBossNotes(UnitName("target"));
		else
			CSR:SendReport();
		end
	end);
	statreport:SetNormalTexture("Interface\\Addons\\ElvUI_LvPlus\\Media\\report");
	statreport:SetHighlightTexture("Interface\\Addons\\ElvUI_LvPlus\\Media\\report-Highlight");
		
	statreport:SetScript("OnEnter", function(self) 
		GameTooltip:SetOwner(self, 'ANCHOR_TOP', 0, 6);
		GameTooltip:AddLine(CSR.L.INFO_DURABILITY_TIP);
		GameTooltip:AddLine(CSR.L.INFO_DURABILITY_TIP2);
		GameTooltip:Show();
	end);
	statreport:SetScript("OnLeave", function(self)
		GameTooltip:Hide();
	end);
end

function CB:Initialize()
	self:LoadChatbar();
	if E.db.LvPlus.LvGeneral.General.LvChatFrame.ChatMSGLoot.EnableBtn then
		self:LoadChatMSGLoot();
	end
	if not E.db.LvPlus.LvGeneral.General.LvChatFrame.SmartChatBub.EnableBtn then
		SetCVar("chatBubbles", 1);
		return
	else
		self:RegisterEvent("PLAYER_ENTERING_WORLD", "ChatBub");
	end
	if E.db.LvPlus.LvGeneral.General.LvChatFrame.ChatBar.ChatBarPoi == 0 then
		self:SecureHook('FCF_SetChatWindowFontSize', "PositionChatbar");
		self:RegisterEvent("PLAYER_ENTERING_WORLD", "PositionChatbar");
		self:RegisterEvent('PET_BATTLE_OPENING_START', "PositionChatbar");
	end
end

local function InitializeCallback()
	if not E.db.LvPlus.LvGeneral.General.LvChatFrame.ChatBar.EnableBtn then
		return;
	end
	CB:Initialize();
end

E:RegisterModule(CB:GetName(), InitializeCallback)